import gzip
import logging
import os.path
import pickle
import re
import sys
import tempfile
import urllib.request

from ..ContentsIndex import ContentsIndex
from ..utils import md5sum, file_type

# TODO: Use xdg.BaseDirectory when ready for python3
CACHE_DIR = os.path.expandvars("$HOME/.cache/archive-cache")


class ArchiveCache:
    """
    A data access object for one release in a Debian FTP repository.
    ArchiveCache allows retrieval of the release's deb packages with
    dict syntax (eg cache['python3.2']). It also creates a searchable
    contents index for all files available for installation in the
    release, as mapped to the packages that install them.

    It is important to note that ArchiveCache is modal: it only looks at
    one architecture's package set at a time, plus those in 'all'. One
    must explicitly switch with cache.load_arch('arch') to inspect
    another architecture's packages.
    """

    def __init__(self, release_uri, index_prefixes=[], memory=500):
        """
        @release_uri: a path (URL or file) like 'dists/lucid/Release'.
        @index_prefixes: a list of directories (like '/usr/share/icons')
            to add as search prefixes in the cache contents index.
        @memory: the maximum amount of memory to index (in MB).
        """
        self._contents_index = None
        self._active_arch = None
        self.index_prefixes = index_prefixes
        self.memory = memory
        self._set_locations(release_uri)
        self._read_release_metadata()
        self._read_package_list()

    def load_arch(self, arch):
        """
        Switches the cache to the given architecture, loading a new
        contents index and new package data as apporpriate. The given
        architecture must be in cache.release['Architectures'].
        """
        if self._contents_index:
            del self._contents_index
        contents = "Contents-{0}.gz".format(arch)
        self._active_arch = arch
        contents_uri = os.path.join(self.suite_uri, contents)
        # get unique identification of the most recent contents archive
        if os.path.exists(contents_uri):
            id_tag = os.stat(contents_uri).st_mtime
            contents = open(contents_uri, "rb")
        else:
            contents = urllib.request.urlopen(contents_uri)
            id_tag = contents.info()['etag'].strip("\"")
        cache_file = "{0}_Contents-{1}.gz".format(id_tag, arch)
        cache_contents = os.path.join(self.suite_dir, cache_file)
        # read contents into cache and delete old cache, if needed
        if not os.path.exists(cache_contents):
            is_cache = lambda f: re.match("_{0}i$".format(contents), f)
            for path in [f for f in os.listdir(self.suite_dir) if is_cache(f)]:
                path = os.path.join(self.suite_dir, path)
                os.remove(path)
            with open(cache_contents, "wb") as f:
                f.write(contents.read())
        contents.close()
        assert file_type(cache_contents) == "gzip compressed data"
        self._contents_index = ContentsIndex(cache_contents,
                                             self.index_prefixes, self.memory)

    def get_info(self, package_name):
        """
        Returns a dictionary of keyword-value metadata pairs for a
        package in the current cache architecture.
        """
        if package_name in self.info[self._active_arch]:
            info_cache = self.info[self._active_arch][package_name]
        elif package_name in self.info["all"]:
            info_cache = self.info["all"][package_name]
        else:
            logging.warning("Package {0} not in current arch: {1}.".format(
                            package_name, self._active_arch))
            return None
        with open(info_cache, "rb") as info_file:
            return pickle.load(info_file)

    def get_component(self, package_name):
        """
        Returns the component name (eg 'universe') that the given
        package name belongs to in the current cache architecture.
        """
        pkg_info = self.get_info(package_name)
        if pkg_info:
            return pkg_info["Component"]

    def __getitem__(self, package_name):
        """
        Returns a file path to this archive's latest deb for the given
        package name (architecture-specific). Returns None if no deb
        can be found [KeyError will not be raised].
        """
        logging.debug("Package {0} in {1} requested from cache.".format(
                      package_name, self._active_arch))
        pkg_info = self.get_info(package_name)
        if not pkg_info:
            return None
        # get from cache, downloading if necessary
        path = pkg_info["Filename"]
        path_uri = os.path.join(self.archive_uri, path)
        path_cache = os.path.join(self.cache_dir, path)
        if not os.path.exists(path_cache):
            logging.debug("Downloading {0} to {1}.".format(
                          path_uri, path_cache))
            path_dir = os.path.dirname(path_cache)
            if not os.path.exists(path_dir):
                os.makedirs(path_dir)
            urllib.request.urlretrieve(path_uri, path_cache)
        if not os.path.exists(path_cache):
            logging.warning("Failed to download {0}".format(path_uri))
        else:
            return path_cache

    def _set_locations(self, release_uri):
        # get useful paths from 'archive/dists/suite/Release'
        self.release_uri = release_uri
        self.suite_uri = os.path.dirname(release_uri)
        self.dists_uri, self.suite = os.path.split(self.suite_uri)
        self.archive_uri = os.path.dirname(self.dists_uri)
        # set local cache directories
        cache_name = self.archive_uri.strip("/").replace("/", "_")
        self.cache_dir = os.path.join(CACHE_DIR, cache_name)
        self.suite_dir = os.path.join(self.cache_dir, "dists", self.suite)
        self.release_file = os.path.join(self.suite_dir, "Release")
        # and download the latest Release file
        logging.info("Loading release metadata at {0}".format(release_uri))
        if not os.path.isdir(self.suite_dir):
            os.makedirs(self.suite_dir)
        urllib.request.urlretrieve(release_uri, self.release_file)
        # finally, if the archive is local, create a pool directory symlink.
        # while we have to copy metadata to ensure it's stable throughout
        # execution, there's no need to wantonly copy the deb files
        if os.path.exists(self.archive_uri):
            logging.info("Using local pool of packages")
            pool_dir = os.path.join(self.cache_dir, "pool")
            local_pool = os.path.join(self.archive_uri, "pool")
            if not os.path.exists(local_pool):
                os.symlink(local_pool, pool_dir)

    def _read_release_metadata(self):
        # get all lines in the Release file and collect key value pairs
        with open(self.release_file, "rb") as f:
            lines = [line.decode() for line in f]
        self._release = {}
        for i, line in enumerate(lines):
            key, value = line.split(":", 1)
            if key == "MD5Sum":
                break
            self.release[key.strip()] = value.strip()
        # read MD5 checksums for all component/binary-*/Packages.gz files
        archs = self.release["Architectures"].split()
        comps = self.release["Components"].split()
        regex = ".+\W({0})/binary-({1})/Packages.gz".format("|".join(comps),
                                                            "|".join(archs))
        self.package_files = {}
        for line in lines[i + 1:]:
            if re.match(regex, line):
                checksum, size, path = line.split()
                self.package_files[path] = (size, checksum)
            # stop on new key value pair (probably SHA1 sums)
            if ":" in line:
                return

    def _read_package_list(self):
        self.info = {}
        for path in self.package_files:
            logging.info("Loading packages metadata at {0}".format(path))
            # get component from path (eg main/binary-i386/Packages.gz)
            component = path.split("/")[0]
            path_file = os.path.join(self.suite_dir, path)

            # ensure that we have the latest Packages.gz cached
            if (not os.path.exists(path_file) or not os.path.getsize(path_file)
                or md5sum(path_file) != self.package_files[path][1]):
                path_uri = os.path.join(self.suite_uri, path)
                path_dir = os.path.dirname(path_file)
                if not os.path.exists(path_dir):
                    os.makedirs(path_dir)
                try:
                    urllib.request.urlretrieve(path_uri, path_file)
                except ValueError:
                    # this Packages.gz does not exist
                    logging.debug("Could not secure {0}".format(path))
                    continue
                if file_type(path_file) != "gzip compressed data":
                    logging.debug("Could not secure {0}".format(path))
                    continue
                assert md5sum(path_file) == self.package_files[path][1]

            # now that we have the list cached, get its package data pairs
            with gzip.open(path_file, "r") as archive:
                data = archive.read().decode()
            sections = data.strip().split("\n\n")
            for section in sections:
                lines = section.strip().split("\n")
                keywords = {}
                last_line_keyword = ""
                for line in lines:
                    if line.startswith(" "):
                        keywords[last_line_keyword] += line
                    else:
                        keyword, value = line.split(":", 1)
                        keywords[keyword.strip()] = value.strip()
                        last_line_keyword = keyword
                pkgname = keywords["Package"]
                arch = keywords["Architecture"]
                keywords["Component"] = component
                if arch not in self.info:
                    self.info[arch] = {}
                info_cache = tempfile.NamedTemporaryFile(delete=False)
                pickle.dump(keywords, info_cache)
                self.info[arch][pkgname] = info_cache.name
                info_cache.close()

    def __del__(self):
        # clean up temporary package data caches
        for arch_pkg in self.info.values():
            for cache_file in arch_pkg.values():
                logging.debug("Removing cache {0}".format(cache_file))
                os.remove(cache_file)

    architecture = property(lambda self: self._active_arch, doc="""
        A string (eg 'i386') indicating the architecture the cache
        currently has loaded or None. Different architectures provide
        different packages for cache['pkgname'] deb retrieval.""")

    contents_index = property(lambda self: self._contents_index, doc="""
        A ContentsIndex instance, allowing search of all file paths
        installable in this release, mapping each to a particular
        package name. Indexes are architecture-specific.""")

    release = property(lambda self: self._release, doc="""
        A dict of all information of the cached released. It likely has
        the version, available components and architectures, and the
        last modification date.""")
