import logging
import os.path
import subprocess
import tarfile
import tempfile
import time

from collections import namedtuple


def md5sum(path):
    """Read the MD5 checksum of a file."""
    results = subprocess.check_output(["md5sum", path]).decode()
    return results.split()[0].strip()


def file_type(path):
    """Read the mimetype of a file."""
    results = subprocess.check_output(["file", "-b", path]).decode()
    return results.split(",")[0].strip()


def file_charset(path):
    """Read the charset of a text/plain file."""
    results = subprocess.check_output(["file", "-ib", path]).decode()
    return results.partition("=")[2].strip()


def read_config(config_dir, config_file):
    """Read all lines from a config file."""
    config_lines = []
    # stop if the file doesn't exist
    path = os.path.join(config_dir, config_file)
    if not os.path.exists(path):
        return config_lines
    logging.info("using config: {0}".format(config_file))
    config = open(path, "rb")
    for line in [line.decode() for line in config]:
        # remove comments and whitespace, and then add
        line = line.split("#")[0].strip()
        if line != "":
            config_lines.append(line)
    config.close()
    return config_lines


# From here on, it's all deb extraction. Rather complicated. The goal is
# to write an extracted 'data.tar.[compression]' in the cache directory,
# open it, and return a member from it.
def load_deb_file(path, deb):
    """
    Returns an open reference to the contents of a file in a deb archive.

    @path: An absolute path for a file within the deb package.
    @deb: The file path to that deb package.
    """
    deb = os.path.abspath(deb)
    cache_dir, deb_file = os.path.split(deb)
    # preserve wokring direcotry so we can return to it
    start_dir = os.getcwd()
    # first try to extract data.tar itself
    try:
        # list all paths in the deb (a deb is simply an ar archive)
        files = subprocess.check_output(["ar", "t", deb]).decode().split()
        # uniquely identify the deb by its contents
        checksum = _checksum_deb(cache_dir, deb, files)
        # extract data.tar or fetch it from that checksum's cache
        data_tar = _extract_data_tar(cache_dir, deb, files, checksum)
    except Exception as e:
        logging.warning("Could not extract {0}: {1}".format(deb, e))
        os.chdir(start_dir)
        return None

    # now try to get load the requested member from data.tar
    data_tarfile = None
    linkless_path = path
    try:
        data_tarfile = tarfile.open(data_tar)
        # all deb data.tar members start with "."
        path = os.path.join(".", path)
        if path.startswith("/"):
            path = ".{0}".format(path)
        # acutally read file, making sure not to trip over symlinks
        linkless_path = _read_symbolic_path(data_tarfile, path)
        data_file = data_tarfile.extractfile(linkless_path)
        temp = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        temp.write(data_file.read())
        temp.close()
    except IOError as e:
        logging.warning("The debfile {0} is not readable: {2}".format(deb, e))
        os.chdir(start_dir)
        return None
    except KeyError as e:
        if linkless_path != path:
            # _read_symbolic_path() changed the extract target,
            # this may mean the target is in another package.
            logging.debug("Encountered foreign symlink {0} in {1}: {2}".format(
                          linkless_path, deb, e))
            os.chdir(start_dir)
            if len(linkless_path):
                return linkless_path[1:]
        logging.debug("{0} is not a member of {1}: {2}".format(path, deb, e))
        return path
    except Exception as e:
        logging.warning("No path {0} in {1}: {2}".format(path, deb, e))
        os.chdir(start_dir)
        if data_tarfile:
            data_tarfile.close()
        return None
    os.chdir(start_dir)
    return open(temp.name, "rb")


def _checksum_deb(path, debpath, files):
    # checksum the deb control file; unless the world is ending, it
    # it should uniquely (and cheaply) identify the package.
    control_file = None
    for f in files:
        if f.startswith("control"):
            control_file = f
            break
    if not control_file:
        raise Exception("Invalid deb {0}: no control file".format(debpath))
    os.chdir("/tmp")
    subprocess.call(["ar", "x", debpath, control_file])
    results = subprocess.check_output(["md5sum", control_file])
    checksum = results.split()[0].strip()
    os.remove(control_file)
    return checksum


def _extract_data_tar(path, debpath, files, checksum):
    # if we already have data.tar in cache, return that
    cache_data_tar = os.path.join(path, "{0}.cache".format(checksum.decode()))
    if os.path.exists(cache_data_tar):
        return cache_data_tar
    # else search the ar archive members for it
    data_file = None
    for f in files:
        if f.startswith("data"):
            data_file = f
            break
    if not data_file:
        raise Exception("Invalid deb {0}: no data file".format(debpath))
    # acutally extract data.tar, cache it, and return the cache path
    os.chdir(path)
    subprocess.call(["ar", "x", debpath, data_file])
    extracted_data = os.path.join(path, data_file)
    if data_file.endswith("lzma"):
        # tarfile does not support lzma. uncompress
        subprocess.call(["lzma", "-d", extracted_data])
        extracted_data = os.path.splitext(extracted_data)[0]
    os.rename(extracted_data, cache_data_tar)
    return cache_data_tar


def _read_symbolic_path(tarfile, extract_path):
    member_paths = [path.rstrip("/") for path in tarfile.getnames()]
    return_path = _ReturnPath(extract_path)
    contains_symlink = False
    start_time = time.time()
    while return_path.remaining:
        next_element = return_path.remaining.pop(0)
        next_path = os.path.join(return_path.done, next_element)
        logging.debug("reading symbolic path: {0}".format(next_path))
        if next_path not in member_paths:
            if contains_symlink:
                if not return_path.remaining:
                    return next_path
                return os.path.join(next_path, "/".join(return_path.remaining))
            raise Exception("tar does not contain {0}".format(next_path))
        # now see if it's symbolic
        next_member = tarfile.getmember(next_path)
        if next_member.issym():
            contains_symlink = True
            # if it's an absolute symbolic link, replace the return path
            if next_member.linkname.startswith("/"):
                return_path.done = ".{0}".format(next_member.linkname)
                continue
            # else discard the new element, and add its linked path
            next_path = os.path.join(return_path.done, next_member.linkname)
            # now resolve all '..' references
            next_path_stack = []
            for element in next_path.split('/'):
                if element == '..':
                    next_path_stack.pop()
                else:
                    next_path_stack.append(element)
            next_path = '/'.join(next_path_stack)
            # ensure we are not caught in an infinite symlink cycle
            if (time.time() - start_time) > 2:
                logging.warn('symlink cycle in %s' % tarfile.name)
                return extract_path
        # update the return path
        return_path.done = next_path
    return return_path.done


class _ReturnPath:
    def __init__(self, extract_path):
        self.remaining = extract_path.split("/")
        self.done =""
