#!/usr/bin/env python3
import os.path
import hashlib
import subprocess
import sys
import unittest
sys.path.insert(0, "..")

from src.ArchiveCache import ArchiveCache
from src.ContentsIndex import ContentsIndex

RELEASE_DATA = {
    "Origin": "ArchiveCache",
    "Suite": "test",
    "Date": "Fri, 10 Jun 2011",
    "Architectures": "foo bar baz",
    "Components": "main alt",
    "Description": "ArchiveCache Test Suite"}

STANDARD_APP_DATA = {
    "Package": "standard-app",
    "Component": "main",
    "Architecture": "foo",
    "Filename": "pool/main/standard/standard-app_1.0-1_foo.deb",
    "MD5sum": "a94aadee76c8ba9d2e843136d2e57849"}

# TODO: replace with xdg_cache_home when python3-xdg is ready
CACHE_DIR = os.path.expandvars("$HOME/.cache/archive-cache")
RELEASE_FILE = "archive/dists/test/Release"
STANDARD_PKG = "pool/main/standard/standard-app_1.0-1_foo.deb"
STANDARD_PKG_MD5 = "a94aadee76c8ba9d2e843136d2e57849"


class TestArchiveCache(unittest.TestCase):
    """
    Tests the proper operation of ArchiveCache, which transparently
    handles the searching and fetching of packages from an archive
    mirror.
    """
    @classmethod
    def setUpClass(self):
        self.cache = ArchiveCache(RELEASE_FILE)

    def test_release_data(self):
        """archive cache should accurately report release information"""
        self.assertEqual(RELEASE_DATA, self.cache.release)

    def test_package_fetching(self):
        """archive cache should allow arch-specific package retrieval"""
        pkg_path = os.path.join(CACHE_DIR, "archive", STANDARD_PKG)
        # provide a path to the deb for a given package name
        self.cache.load_arch("foo")
        self.assertEqual(self.cache["standard-app"], pkg_path)
        with open(pkg_path, "rb") as f:
            md5_checksum = hashlib.md5(f.read()).hexdigest()
        self.assertEqual(md5_checksum, STANDARD_PKG_MD5)
        # do not return any package not available for the current arch
        self.cache.load_arch("bar")
        self.assertEqual(self.cache["standard-app"], None)

    def test_package_component(self):
        """archive cache should provide each package's component"""
        self.cache.load_arch("foo")
        self.assertEqual(self.cache.get_component("standard-app"), "main")
        self.cache.load_arch("bar")
        self.assertEqual(self.cache.get_component("standard-app"), None)

    def test_package_info(self):
        """archive cache should support fetching Packages.gz info"""
        self.cache.load_arch("foo")
        self.assertEqual(self.cache.get_info("standard-app"), STANDARD_APP_DATA)
        self.cache.load_arch("bar")
        self.assertEqual(self.cache.get_info("standard-app"), None)

    def test_searchable(self):
        """archive cache should have indexes for path"""
        self.cache.load_arch("foo")
        self.assertIsInstance(self.cache.contents_index, ContentsIndex)
        # ensure indexes are based of the correct arch's Contents file
        contents_file = self.cache.contents_index.contents_file
        self.assertTrue(contents_file.endswith("foo.gz"))
        self.cache.load_arch("bar")
        contents_file = self.cache.contents_index.contents_file
        self.assertTrue(contents_file.endswith("bar.gz"))


if __name__ == "__main__":
    subprocess.call("./build_test_archive")
    unittest.main()
